/*
 * Generic pager functionality.
 *
 * Copyright (C) 2021, 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include "dataspace_server.h"
#include "ipc.h"
#include "pager.h"

#include "copied_page_mapper.h"
#include "masked_page_mapper.h"



/* Initialise the pager with a page mapper and the given flags controlling
   access to a file. */

Pager::Pager(GenericPageMapper *mapper, map_flags_t map_flags)
: _start(0), _size(0),
  _mapper(mapper), _mapper_base(mapper), _mapper_masked(NULL),
  _mapper_copied(NULL),
  _map_flags(map_flags)
{
}

/* Close the pager. */

void Pager::close()
{
    if (_mapper_masked != NULL)
    {
        delete _mapper_masked;
        _mapper_masked = NULL;
    }

    if (_mapper_copied != NULL)
    {
        delete _mapper_copied;
        _mapper_copied = NULL;
    }
}

/* Flush data to the file. */

long Pager::flush(offset_t populated_size, offset_t *size)
{
    _mapper->flush_all(_start, populated_size);

    *size = _mapper->get_data_size();
    return L4_EOK;
}

/* Resize the underlying file. */

long Pager::resize(offset_t *size)
{
    _mapper->set_data_size(*size);

    *size = _mapper->get_data_size();
    return L4_EOK;
}

/* Expose a region of the file. */

long Pager::mmap(offset_t position, offset_t length,
                 offset_t start_visible, offset_t end_visible,
                 offset_t *start_pos, offset_t *end_pos, offset_t *size)
{
    /* Define region characteristics. */

    _start = trunc(position, PAGE_SIZE);
    _size = round(position + length, PAGE_SIZE) - _start;

    /* Return the start and end positions plus populated extent. */

    *start_pos = _start;
    *end_pos = _start + _size;
    *size = _mapper_base->get_data_size();

    /* Permit masking of mapped regions. */

    if ((start_visible || end_visible) &&
        ((*start_pos != start_visible) || (*end_pos != end_visible)))
    {
        /* Introduce the masked page and copied page mappers. */

        _mapper_masked = new MaskedPageMapper(_mapper_base, start_visible, end_visible);
        _mapper_copied = new CopiedPageMapper(_mapper_masked);
        _mapper = _mapper_copied;
    }
    else
        _mapper = _mapper_base;

    return L4_EOK;
}

/* Map a flexpage corresponding to the dataspace 'offset' involving a 'hot_spot'
   (flexpage offset). */

long Pager::map(offset_t offset, address_t hot_spot, map_flags_t map_flags,
                l4_snd_fpage_t *region)
{
    offset_t file_offset = _start + offset;
    offset_t max_offset = _start + _size;

    /* Prevent access beyond that defined by the pager.
       NOTE: Permitting executable requests here. This needs to be configured
             when opening the pager or by another means. */

    if (map_flags & (~(_map_flags | L4RE_DS_F_X)))
        return -L4_EACCESS;

    /* Obtain a flexpage from the page mapper. */

    Flexpage *flexpage = _mapper->get(file_offset, map_flags);

    /* Issue the flexpage via the IPC system. */

    long err = ipc_prepare_flexpage(flexpage, file_offset, max_offset,
                                    hot_spot, map_flags, region);

    if (!err)
        err = complete_Dataspace_map(*region);

    /* After the obtained flexpage is issued, it is queued for future reuse, if
       appropriate for the mapper concerned. */

    _mapper->queue(flexpage);

    if (err)
        return err;

    return IPC_MESSAGE_SENT;
}

/* Return the total size of the data. */

offset_t Pager::get_data_size()
{
    return _mapper->get_data_size();
}

// vim: tabstop=4 expandtab shiftwidth=4
